(*
    This file is part of the Geometry project -
    which provides some useful geometry stuff.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

functor FIRect2D() :> IRECT2D =
  struct
    (* Public types and values *)
    type t = {x : int, y : int, w : int, h : int}

    val zeroValue = {x = 0, y = 0, w = 0, h = 0}

    (* Private stuff *)
    fun contains1D {min, max, check} =
          check >= min andalso check < max (* Notice width... *)

    fun max (a, b) =
          if a > b then a else b

    fun min (a, b) =
          if a < b then a else b

    fun intersect1D {p1, s1, p2, s2} =
          {min = max (p1, p2),
           max = min (p1 + s1, p2 + s2)}

    fun union1D {p1, s1, p2, s2} =
          {min = min (p1, p2),
           max = max (p1 + s1, p2 + s2)}

    (* Public functions *)
    (* FIXME: Rename to hasNoArea? or isEmpty? or isDegenerate? *)
    fun isZero ({x, y, w, h} : t) =
          w <= 0 orelse h <= 0

    fun equal ({x, y, w, h} : t,
               {x = x', y = y', w = w', h = h'} : t) =
          x = x' andalso y = y' andalso
          w = w' andalso h = h'

    (* If the rectangle gets zero extent - zeroValue is returned *)
    fun inset ({x, y, w, h} : t, amount) =
          let
            val result =
                  {x = x + amount, y = y + amount,
                   w = w - (amount * 2), h = h - (amount * 2)}
          in
            if isZero result then
              zeroValue
            else
              result
          end

    fun move ({x, y, w, h} : t, (dx, dy)) =
          {x = x + dx, y = y + dy, w = w, h = h}

    (* A point on the boudary of a rectangle is also
       considered to be contained in the rectangle. *)
    fun containsPoint ({x, y, w, h} : t, {x = px, y = py}) =
          contains1D {min = x, max = x + w, check = px} andalso
          contains1D {min = y, max = y + h, check = py}

    (* Returns zeroValue for empty intersections
       i.e. intersections with zero or less height or width. *)
    fun intersect ({x, y, w, h} : t,
                   {x = x2, y = y2, w = w2, h = h2} : t) =
          let
            val {min = minx, max = maxx} =
                   intersect1D {p1 = x, s1 = w, p2 = x2, s2 = w2}
            val {min = miny, max = maxy} =
                   intersect1D {p1 = y, s1 = h, p2 = y2, s2 = h2}

            val width = maxx - minx
            val height = maxy - miny
          in
            if width <= 0 orelse height <= 0 then
              zeroValue
            else
              {x = minx, y = miny, w = width, h = height}
          end

    fun union ((r1 as {x, y, w, h} : t),
               (r2 as {x = x2, y = y2, w = w2, h = h2} : t)) =
          if isZero r1 then
            if isZero r2 then
              zeroValue
            else r2
          else if isZero r2 then
            r1
          else
            let
              val {min = minx, max = maxx} =
                     union1D {p1 = x, s1 = w, p2 = x2, s2 = w2}
              val {min = miny, max = maxy} =
                     union1D {p1 = y, s1 = h, p2 = y2, s2 = h2}
           
              val width = maxx - minx
              val height = maxy - miny
            in
              {x = minx, y = miny, w = width, h = height}
            end
  end

structure IRect2D = FIRect2D()