(*
    This file is part of the Geometry project -
    which provides some useful geometry stuff.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)


structure Vector3D : VECTOR3D where type coord = real =
  struct
    structure Math : MATH = Math

    type coord = real
    type scalar = real
    type t = {x : coord, y : coord, z : coord}
    type vertex = t

    exception Divide (* FIXME: Should be a general exception or a general error handler *)

    (* Methods for addition arithmetics *)
    val zeroValue =
          {x = 0.0, y = 0.0, z = 0.0}

    fun negate ({x, y, z} : t) =
          {x = ~x, y = ~y, z = ~z}

    fun add ({x = x0, y = y0, z = z0} : t,
             {x = x1, y = y1, z = z1} : t) =
          {x = x0 + x1, y = y0 + y1, z = z0 + z1}

    fun subtract ({x = x0, y = y0, z = z0} : t,
                  {x = x1, y = y1, z = z1} : t) =
          {x = x0 - x1, y = y0 - y1, z = z0 - z1}

    (* Scaling *)
    fun scale ({x, y, z} : t, factor : real) =
          {x = x * factor,
           y = y * factor,
           z = z * factor}

    fun inverseScale ({x, y, z} : t, divisor : real) =
          if Real.==(0.0, divisor) then
            raise Divide
          else
            {x = x / divisor,
             y = y / divisor,
             z = z / divisor}

    (* Comparison *)
    fun equal ({x = x0, y = y0, z = z0} : t,
               {x = x1, y = y1, z = z1} : t) =
          (Real.==(x0, x1) andalso
           Real.==(y0, y1) andalso
           Real.==(z0, z1))

    (* Special 3D vector methods... *)
    fun fromVertices({x = x0, y = y0, z = z0} : vertex,
                     {x = x1, y = y1, z = z1} : vertex) =
          {x = x1 - x0, y = y1 - y0, z = z1 - z0}
  
    fun dotProduct({x = x0, y = y0, z = z0} : t,
                   {x = x1, y = y1, z = z1} : t) =
          (x0 * x1) + (y0 * y1) + (z0*z1)

    fun crossProduct({x = x0, y = y0, z = z0} : t,
                     {x = x1, y = y1, z = z1} : t) =
          {x = (y0 * z1) - (z0 * y1),
           y = (z0 * x1) - (x0 * z1),
           z = (x0 * y1) - (y0 * x1)}

    fun length ({x, y, z} : t) =
          Math.sqrt ((x * x) + (y * y) + (z * z))
    
    fun normalize (vec as {x, y, z} : t) =
          let
            val vecLen = length vec
          in
            if Real.==(vecLen, 0.0) then
              raise Divide
            else
              {x = x / vecLen,
               y = y / vecLen,
               z = z / vecLen}
          end
  end
